#include <math.h>
#include <stdio.h>
#include "fixdefs.h"
#include "matrix.h"

int minv (vektor mp[], int p)
/* invertiert pxp-Matrix mp (0, wenn nicht invertierbar) */
/* Benoetigt Funktion vg */
/* Benoetigt PKOR */
{
  int i, j, k, l, invcon;
  vektor r1, s1, pm[PKOR];
  double co;

  invcon=1;
/* TZ printf("  Beginn minv\n"); */
  for (i=0; i<p; ++i)
    for (j=0; j<p; ++j)
    {
      if (i==j)
		pm[i][j]=1;
	  else
		pm[i][j]=0;
	 }
  for (i=0; i<p-1; ++i)
  {
    k=i+1;
    while ((fabs(mp[i][i])<1e-9) && invcon)
	{
	  if (k==p)
	    invcon=0;
	  vg(r1, mp[i], p); vg(s1, pm[i], p);
	  vg(mp[i], mp[k], p); vg(pm[i], pm[k], p);
	  vg(mp[k], r1, p); vg(pm[k], s1, p);
	  k++;
	}
	if (!invcon)
	  i=p-1;
	for (j=i+1; j<p; ++j)
	{
	  co=mp[j][i]/mp[i][i];
	  for (l=0; l<p; ++l)
	  {
		  mp[j][l]=mp[j][l]-co*mp[i][l];
		  pm[j][l]=pm[j][l]-co*pm[i][l];
	  }
	}
  }
  for (i=p-1; i>=1; --i)
  {
    if ((fabs(mp[i][i])>=1e-9) && invcon)
    {
      for (j=i-1; j>=0; --j)
      {
	co=mp[j][i]/mp[i][i];
	for (l=0; l<p; ++l)
	  pm[j][l]=pm[j][l]-co*pm[i][l];
      }
    }
	 else
      invcon=0;
  }
  for (i=0; i<p; ++i)
  {
    if ((fabs(mp[i][i])>=1e-9) && invcon)
    {
      for (j=0; j<p; ++j)
	pm[i][j]=pm[i][j]/mp[i][i];
    }
    else
      invcon=0;
  }
  for (i=0; i<p; ++i)
	vg(mp[i], pm[i], p);
/*  if (!invcon)
    printf("Warning! Matrix not numerically invertable. \n"); */
  return(invcon);
/* TZ printf("  Ende minv\n");a */
}

void vg (double v1[], double v2[], int dim)
/* Setzt Vektor v1=Vektor v2 */
{
  int i;

  for(i=0; i<dim; ++i)
    v1[i]=v2[i];
}

void mg (vektor m1[], vektor m2[], int p, int q)
/* Setzt pxq-Matrix m1=m2 */
/* Benoetigt PKOR */
{
  int i,j;

  for(i=0; i<p; ++i)
    for(j=0; j<q; ++j)
      m1[i][j]=m2[i][j];
}

double vmult (double v1[], double v2[], int p)
/* v1 transponiert*v2 */
{
  int i;
  double su=0;

  for (i=0; i<p; ++i)
    su+=v1[i]*v2[i];
  return (su);
}

void matmult (vektor m1[], vektor m2[], vektor m3[], int p, int q, int r)
/* m3=m1*m2 (m1 pxq, m2 qxr) */
/* Benoetigt PKOR */
{
  int i, j, k;

  for (i=0; i<p; ++i)
    for (k=0; k<r; ++k)
    {
      m3[i][k]=0;
      for (j=0; j<q; ++j)
        m3[i][k]+=m1[i][j]*m2[j][k];
    }
}

void vmmult (double v1[], vektor m2[], double v3[], int p, int q)
/* v3=v1'*m2, m2 p*q */
/* Benoetigt PKOR */
{
  int i,j;

  for (i=0; i<q; ++i)
  {
    v3[i]=0;
    for (j=0; j<p; ++j)
      v3[i]+=v1[j]*m2[j][i];
  }
}

void jacobi (vektor m1[], vektor m2[], int p)
/* Eigenwert-Zerlegung von m1 (symm.), spaeter m1: Diagonale, m2: Eigenvektoren */
/* Nach Schwarz e.a. Matrix-Numerik, S. 113 ff. */
/* Benoetigt Funktionen rotation, matmult, mg */
/* Benoetigt PKOR */
{
  int i,j,k,l;
  vektor m3[PKOR], m4[PKOR];
  double max, theta, t, s, c;

  for (i=0; i<p; ++i)
  {
    for (j=0; j<p; ++j)
    {
      if (i==j)
        m2[i][j]=1;
      else
        m2[i][j]=0;
    } /* for j */
  } /* for i */
  do
  {
    max=0;
    for (i=0; i<p-1; ++i)
    {
      for (j=i+1; j<p; ++j)
      {
        if (fabs(m1[i][j])>max)
        {
          max=fabs(m1[i][j]); k=i; l=j;
/* TZ  printf ("\n m1[%d][%d]= %6.4lf \n",i,j,m1[i][j]); */
        } /* if */
      } /* for j */
    } /* for i */
/* TZ printf ("\n max= %6.4lf \n",max); */
    if (p*max >= 1e-8)
    {
      theta=0.5*(m1[l][l]-m1[k][k])/m1[k][l];
      if (theta==0)
        t=1;
      else
        t=1/(theta+theta*sqrt(1+theta*theta)/fabs(theta));
      c=1/sqrt(1+t*t); s=c*t;
/* TZ  printf("\nRotation mit k=%d, l=%d, c=%6.4lf s=%6.4lf\n",k,l,c,s); */
/* TZ scanf ("%s"); */
/* TZ printf("\nm1[k][k]=%6.4lf m1[k][l]=%6.4lf m1[l][l]=%6.4lf\n", m1[k][k],m1[k][l], m1[l][l]); */
      rotation (p,c,s,k,l,m1);
/* TZ printf("\nm1[k][k]=%6.4lf m1[k][l]=%6.4lf m1[l][l]=%6.4lf\n", m1[k][k],m1[k][l], m1[l][l]); */
      for (i=0; i<p; ++i)
      {
        for (j=0; j<p; ++j)
        {
          if (i==j)
            m3[i][j]=1;
          else
            m3[i][j]=0;
        } /* for j */
      } /* for i */
      m3[l][l]=m3[k][k]=c; m3[k][l]=-s;
      m3[l][k]=s;
      matmult(m3, m2, m4, p,p,p);
      mg (m2, m4, p ,p);
    } /* if */
  } /* do */
  while (p*max>=1e-8);
}

void rotation (int p, double c, double s, int k, int l, vektor m1[])
/* Jacobi-Rotation der Matrix m1, k, l: Elemente, c, s: Rotationsmatrix */
{
  double g,h;
  int j;

  h=c*c*m1[k][k]-2*c*s*m1[k][l]+s*s*m1[l][l];
  g=s*s*m1[k][k]+2*c*s*m1[k][l]+c*c*m1[l][l];
  m1[l][k]=c*s*(m1[k][k]-m1[l][l])+(c*c-s*s)*m1[k][l];
  m1[k][l]=m1[l][k];
  m1[k][k]=h; m1[l][l]=g;
  for (j=0; j<k; ++j)
  {
    h=c*m1[j][k]-s*m1[j][l];
    m1[j][l]=s*m1[j][k]+c*m1[j][l];
    m1[j][k]=h;
    m1[k][j]=m1[j][k];
    m1[l][j]=m1[j][l];
  }
  for (j=k+1; j<l; ++j)
  {
    h=c*m1[k][j]-s*m1[j][l];
    m1[j][l]=s*m1[k][j]+c*m1[j][l];
    m1[k][j]=h;
    m1[j][k]=m1[k][j];
    m1[l][j]=m1[j][l];
  }
  for (j=l+1; j<p; ++j)
  {
    h=c*m1[k][j]-s*m1[l][j];
    m1[l][j]=s*m1[k][j]+c*m1[l][j];
    m1[k][j]=h;
    m1[j][k]=m1[k][j];
    m1[j][l]=m1[l][j];
  }
}
